import bblfsh


IDENTIFIER = bblfsh.role_id("IDENTIFIER")
QUALIFIED = bblfsh.role_id("QUALIFIED")
LITERAL = bblfsh.role_id("LITERAL")
OPERATOR = bblfsh.role_id("OPERATOR")
EXPRESSION = bblfsh.role_id("EXPRESSION")
LEFT = bblfsh.role_id("LEFT")
BINARY = bblfsh.role_id("BINARY")
ASSIGNMENT = bblfsh.role_id("ASSIGNMENT")
FUNCTION = bblfsh.role_id("FUNCTION")
DECLARATION = bblfsh.role_id("DECLARATION")
NAME = bblfsh.role_id("NAME")
